from sage.all import EllipticCurve
from sage.all import Integer
from sage.arith.misc import inverse_mod
from sage.rings.finite_rings.integer_mod import *
from sage.misc.functional import lift

from montgomery_arithmetics import *
from utilities.discrete_log import *
from utilities.pairing import *
from utilities.supersingular import *


def compressed_pubkey(Alice_pub,b,c,elligator_tables, p,Fp,Fp2):

    
    phiPc = Alice_pub[0]
    phiQc = Alice_pub[1]
    phiPb = Alice_pub[2]
    phiQb = Alice_pub[3]
    phiPQb = Alice_pub[4]
    A3 = get_A(phiPb,phiQb,phiPQb)

    S = MontoEllip(A3,[phiPc,phiQc,phiPb,phiQb,phiPQb])

    PQcheck = S[1][2] + S[1][3]
    if (PQcheck != S[1][4]) and (PQcheck != (-1)*S[1][4]):
        S[1][3] = (-1)*S[1][3]

    c_ex = 0
    check_ex = (p+1)//(2**c)
    while check_ex % 2 == 0:
        c_ex += 1
        check_ex = check_ex // 2

    b_ex = 0
    check_ex = (p+1)//(5**b)
    while check_ex % 5 == 0:
        b_ex += 1
        check_ex = check_ex // 5
    

    P, Q = torsion_basis(S[0], (2**(c+c_ex) * 5**(b+b_ex)))

    P1 = 5**(b+b_ex) * P
    Q1 = 5**(b+b_ex) * Q

    window = [c//10]

    D2 = 2**(c+c_ex)
    e = (p**2 - 1)//D2

    criteria_c = tate_pairing_pari(P1,Q1,D2)**e
    
    phiPce1 = tate_pairing_pari(S[1][0],Q1,D2)**e
    phiPce2 = tate_pairing_pari(P1,S[1][0],D2)**e
    phiQce1 = tate_pairing_pari(S[1][1],Q1,D2)**e
    phiQce2 = tate_pairing_pari(P1,S[1][1],D2)**e

    phiPc_v1 = windowed_pohlig_hellman(phiPce1,criteria_c,c+c_ex,window)//(2**c_ex)
    phiPc_v2 = windowed_pohlig_hellman(phiPce2,criteria_c,c+c_ex,window)//(2**c_ex)
    phiQc_v1 = windowed_pohlig_hellman(phiQce1,criteria_c,c+c_ex,window)//(2**c_ex)
    phiQc_v2 = windowed_pohlig_hellman(phiQce2,criteria_c,c+c_ex,window)//(2**c_ex)

    if phiPc_v1 % 2 == 0:

        phi_inv = inverse_mod(phiPc_v2,2**c)
        
        phiPc_v1 = (phiPc_v1 * phi_inv) % 2**c
        phiQc_v1 = (phiQc_v1 * phi_inv) % 2**c
        phiQc_v2 = (phiQc_v2 * phi_inv) % 2**c
        phiC = [0,phiPc_v1,phiQc_v1,phiQc_v2] 

        
    else:

        phi_inv = inverse_mod(phiPc_v1,2**c)

        phiPc_v2 = (phiPc_v2 * phi_inv) % (2**c)
        phiQc_v1 = (phiQc_v1 * phi_inv) % (2**c)
        phiQc_v2 = (phiQc_v2 * phi_inv) % (2**c)
        phiC = [1,phiPc_v2,phiQc_v1,phiQc_v2]


    
    P2 = D2 * P
    Q2 = D2 * Q

    D5 = 5**(b+b_ex)
    e = (p**2 - 1)//D5
    
    criteria_b = tate_pairing_pari(P2,Q2,D5)** e
    
    phiPbe1 = tate_pairing_pari(S[1][2],Q2,D5)** e
    phiPbe2 = tate_pairing_pari(P2,S[1][2],D5)** e
    phiQbe1 = tate_pairing_pari(S[1][3],Q2,D5)** e
    phiQbe2 = tate_pairing_pari(P2,S[1][3],D5)** e

    phiPb_v1 = discrete_log_pari(phiPbe1,criteria_b,D5)//(5**b_ex)
    phiPb_v2 = discrete_log_pari(phiPbe2,criteria_b,D5)//(5**b_ex)
    phiQb_v1 = discrete_log_pari(phiQbe1,criteria_b,D5)//(5**b_ex)
    phiQb_v2 = discrete_log_pari(phiQbe2,criteria_b,D5)//(5**b_ex)

    if phiPb_v1 % 5 == 0:

        inv_phib = inverse_mod(phiPb_v2,5**b)
        
        phiPb_v1 = (phiPb_v1 * inv_phib) % 5**b
        phiQb_v1 = (phiQb_v1 * inv_phib) % 5**b
        phiQb_v2 = (phiQb_v2 * inv_phib) % 5**b
        
        phiB = [0,phiPb_v1,phiQb_v1,phiQb_v2]
    else:

        inv_phib = inverse_mod(phiPb_v1,5**b)
        
        phiPb_v2 = (phiPb_v2 * inv_phib) % 5**b
        phiQb_v1 = (phiQb_v1 * inv_phib) % 5**b
        phiQb_v2 = (phiQb_v2 * inv_phib) % 5**b
        
        phiB = [1,phiPb_v2,phiQb_v1,phiQb_v2]
    
    return [A3,phiC,phiB]


def decomp_pubkey(Alice_pub_comp,b,c,elligator_tables, p,Fp,Fp2):

    S = MontoEllip(Alice_pub_comp[0],[])

    c_ex = 0
    check_ex = (p+1)//(2**c)
    while check_ex % 2 == 0:
        c_ex += 1
        check_ex = check_ex // 2

    b_ex = 0
    check_ex = (p+1)//(5**b)
    while check_ex % 5 == 0:
        b_ex += 1
        check_ex = check_ex // 5

    P, Q = torsion_basis(S[0], (2**(c+c_ex)) * (5**(b+b_ex)))

    D2 = (2**(c+c_ex)) * (5**b_ex)
    P1 = D2 * P
    Q1 = D2 * Q

    if Alice_pub_comp[2][0] == 0:
        phiPb_point = (Alice_pub_comp[2][1]) * P1 + Q1
        phiQb_point = (Alice_pub_comp[2][2]) * P1 + (Alice_pub_comp[2][3]) * Q1

        phiPQb_point = phiPb_point + phiQb_point

        phiPb = [phiPb_point[0],phiPb_point[2]]
        phiQb = [phiQb_point[0],phiQb_point[2]]
        phiPQb = [phiPQb_point[0],phiPQb_point[2]]

    else:
        phiPb_point = P1 + (Alice_pub_comp[2][1]) * Q1
        phiQb_point = (Alice_pub_comp[2][2]) * P1 + (Alice_pub_comp[2][3]) * Q1

        phiPQb_point = phiPb_point + phiQb_point

        phiPb = [phiPb_point[0],phiPb_point[2]]
        phiQb = [phiQb_point[0],phiQb_point[2]]
        phiPQb = [phiPQb_point[0],phiPQb_point[2]]

    D5 = (5**(b+b_ex)) * (2**c_ex)
    P2 = D5 * P
    Q2 = D5 * Q

    if Alice_pub_comp[1][0] == 0:
        phiPc_point = (Alice_pub_comp[1][1]) * P2 + Q2
        phiQc_point = (Alice_pub_comp[1][2]) * P2 + (Alice_pub_comp[1][3]) * Q2

        phiPc = [phiPc_point[0],phiPc_point[2]]
        phiQc = [phiQc_point[0],phiQc_point[2]]

    else:
        phiPc_point = P2 + (Alice_pub_comp[1][1]) * Q2
        phiQc_point = (Alice_pub_comp[1][2]) * P2 + (Alice_pub_comp[1][3]) * Q2

        phiPc = [phiPc_point[0],phiPc_point[2]]
        phiQc = [phiQc_point[0],phiQc_point[2]]

    return [phiPc,phiQc,phiPb,phiQb,phiPQb]


def compressed_cipher(Bob_pub,c,elligator_tables,p,Fp,Fp2):

    S1 = MontoEllip(Bob_pub[0][0],[Bob_pub[0][1],Bob_pub[0][2]])

    c_ex = 0
    check_ex = (p+1)//(2**c)
    while check_ex % 2 == 0:
        c_ex += 1
        check_ex = check_ex // 2

    D2 = 2**(c+c_ex)
    P1, Q1 = entangled_torsion_basis(S1[0], elligator_tables, (p+1)//D2)


    e = (p**2 - 1)//D2

    criteria_1 = tate_pairing_pari(P1,Q1,D2)** e
    
    phiPce1 = tate_pairing_pari(S1[1][0],Q1,D2)** e
    phiPce2 = tate_pairing_pari(P1,S1[1][0],D2)** e
    phiQce1 = tate_pairing_pari(S1[1][1],Q1,D2)** e
    phiQce2 = tate_pairing_pari(P1,S1[1][1],D2)** e

    window = [c//10]

    phiPc_v1 = windowed_pohlig_hellman(phiPce1,criteria_1,c+c_ex,window)//(2**c_ex)
    phiPc_v2 = windowed_pohlig_hellman(phiPce2,criteria_1,c+c_ex,window)//(2**c_ex)
    phiQc_v1 = windowed_pohlig_hellman(phiQce1,criteria_1,c+c_ex,window)//(2**c_ex)
    phiQc_v2 = windowed_pohlig_hellman(phiQce2,criteria_1,c+c_ex,window)//(2**c_ex)

    if phiPc_v1 % 2 == 0:

        inv_phi = inverse_mod(phiPc_v2,2**c)
        
        phiPc_v1 = (phiPc_v1 * inv_phi) % 2**c
        phiQc_v1 = (phiQc_v1 * inv_phi) % 2**c
        phiQc_v2 = (phiQc_v2 * inv_phi) % 2**c
        
        phiC1 = [0,phiPc_v1,phiQc_v1,phiQc_v2]

    else:

        inv_phi = inverse_mod(phiPc_v1,2**c)
        
        phiPc_v2 = (phiPc_v2 * inv_phi) % 2**c
        phiQc_v1 = (phiQc_v1 * inv_phi) % 2**c
        phiQc_v2 = (phiQc_v2 * inv_phi) % 2**c
        
        phiC1 = [1,phiPc_v2,phiQc_v1,phiQc_v2]

        

    S2 = MontoEllip(Bob_pub[1][0],[Bob_pub[1][1],Bob_pub[1][2]])
    
    P2, Q2 = entangled_torsion_basis(S2[0], elligator_tables, (p+1)//D2)

    criteria_2 = tate_pairing_pari(P2,Q2,D2)**e
    
    phiPce1 = tate_pairing_pari(S2[1][0],Q2,D2)**e
    phiPce2 = tate_pairing_pari(P2,S2[1][0],D2)**e
    phiQce1 = tate_pairing_pari(S2[1][1],Q2,D2)**e
    phiQce2 = tate_pairing_pari(P2,S2[1][1],D2)**e

    phiPc_v1_ = windowed_pohlig_hellman(phiPce1,criteria_2,c+c_ex,window)//(2**c_ex)
    phiPc_v2_ = windowed_pohlig_hellman(phiPce2,criteria_2,c+c_ex,window)//(2**c_ex)
    phiQc_v1_ = windowed_pohlig_hellman(phiQce1,criteria_2,c+c_ex,window)//(2**c_ex)
    phiQc_v2_ = windowed_pohlig_hellman(phiQce2,criteria_2,c+c_ex,window)//(2**c_ex)


    if phiPc_v1_ % 2 == 0:

        inv_phi = inverse_mod(phiPc_v2_,2**c)
        
        phiPc_v1_ = (phiPc_v1_ * inv_phi) % 2**c
        phiQc_v1_ = (phiQc_v1_ * inv_phi) % 2**c
        phiQc_v2_ = (phiQc_v2_ * inv_phi) % 2**c
        
        phiC2 = [0,phiPc_v1_,phiQc_v1_,phiQc_v2_]

    else:

        inv_phi = inverse_mod(phiPc_v1_,2**c)
        
        phiPc_v2_ = (phiPc_v2_ * inv_phi) % 2**c
        phiQc_v1_ = (phiQc_v1_ * inv_phi) % 2**c
        phiQc_v2_ = (phiQc_v2_ * inv_phi) % 2**c
        
        phiC2 = [1,phiPc_v2_,phiQc_v1_,phiQc_v2_]

    return [[Bob_pub[0][0],phiC1],[Bob_pub[1][0],phiC2]]


def decomp_cipher(Bob_pub_comp,Alice_secret,a,b,c,ell,elligator_tables,p,Fp,Fp2):

    S1 = MontoEllip(Bob_pub_comp[0][0],[])

    c_ex = 0
    check_ex = (p+1)//(2**c)
    while check_ex % 2 == 0:
        c_ex += 1
        check_ex = check_ex // 2

    D2 = 2**(c+c_ex)
    P1, Q1 = entangled_torsion_basis(S1[0], elligator_tables, (p+1)//D2)
    P1 = (2**c_ex) * P1
    Q1 = (2**c_ex) * Q1

    if Bob_pub_comp[0][1][0] == 0:

        tauPc_point = (Bob_pub_comp[0][1][1]) * P1 + Q1
        tauQc_point = (Bob_pub_comp[0][1][2]) * P1 + (Bob_pub_comp[0][1][3]) * Q1

    else:

        tauPc_point = P1 + (Bob_pub_comp[0][1][1]) * Q1
        tauQc_point = (Bob_pub_comp[0][1][2]) * P1 + (Bob_pub_comp[0][1][3]) * Q1


    S2 = MontoEllip(Bob_pub_comp[1][0],[])

    P2, Q2 = entangled_torsion_basis(S2[0], elligator_tables, (p+1)//D2)
    P2 = (2**c_ex) * P2
    Q2 = (2**c_ex) * Q2

    if Bob_pub_comp[1][1][0] == 0:

        phiPc_point = (Bob_pub_comp[1][1][1]) * P2 + Q2
        phiQc_point = (Bob_pub_comp[1][1][2]) * P2 + (Bob_pub_comp[1][1][3]) * Q2

    else:

        phiPc_point = P2 + (Bob_pub_comp[1][1][1]) * Q2
        phiQc_point = (Bob_pub_comp[1][1][2]) * P2 + (Bob_pub_comp[1][1][3]) * Q2


    tauPc_point = Alice_secret[0] * tauPc_point
    tauQc_point = Alice_secret[1] * tauQc_point

    value_tau = (weil_pairing_pari(tauPc_point,tauQc_point,2**c))**(3**a)
    value_phi = (weil_pairing_pari(phiPc_point,phiQc_point,2**c))**((2**c - 3**a)//ell)
    

    window = [c//10]
    c02 = windowed_pohlig_hellman(value_tau,value_phi,c,window)

    if not mod(c02,2**c).is_square():
        c02 = (-1)*c02

    c0 = mod(c02,2**c).sqrt()

    phiPc_point = (Alice_secret[0] * Integer(c0)) * phiPc_point
    phiQc_point = (Alice_secret[1] * Integer(c0)) * phiQc_point

    tauPc = [tauPc_point[0],tauPc_point[2]]
    tauQc = [tauQc_point[0],tauQc_point[2]]
    phiPc = [phiPc_point[0],phiPc_point[2]]
    phiQc = [phiQc_point[0],phiQc_point[2]]

    return [[Bob_pub_comp[0][0],tauPc,tauQc],[Bob_pub_comp[1][0],phiPc,phiQc]]
