from random import randrange
from sage.all import EllipticCurve

from montgomery_arithmetics import *
from richelot_isogeny.richelot_isogenies import *

def Alice_keygen(c,a,pub_param,p,Fp,Fp2):
    
    Pc = pub_param[0][0]
    Qc = pub_param[0][1]
    Pb = pub_param[0][2]
    Qb = pub_param[0][3]
    PQb = pub_param[0][4]
    A1 = get_A(Pb,Qb,PQb)

    Alice_secret = [randrange(2**(c-1)),randrange(2**(c-1))]
    Alice_secret[0] = 2*Alice_secret[0]+1
    Alice_secret[1] = 2*Alice_secret[1]+1

    Pc = scalar(Pc,A1,Alice_secret[0])
    Qc = scalar(Qc,A1,Alice_secret[1])

    Alice_pub = [Pc,Qc,Pb,Qb,PQb]

    kernelP = [1,0]
    while kernelP[1] == 0:
        xP = Fp2.random_element()
        kernelP = scalar([xP,1],A1,(p+1)//3)
        if scalar(kernelP,A1,3)[1] != 0:
            kernelP[1] = 0

    kernel = kernelP[0]/kernelP[1]

    zeta3 = 1
    while zeta3 == 1:
        zeta3 = (Fp2.random_element())**((p**2-1)//3)

    for i in range(a-1):
        Alice_pub,kernel = isog_3_random(Alice_pub,kernel,zeta3,p)

    for i in range(len(Alice_pub)):
        Alice_pub[i] = isog_point(Alice_pub[i],[[kernel,1]])

    return Alice_pub,Alice_secret



def Bob_keygen(c,b,pub_param,Alice_pub,strategy,p,Fp,Fp2):
    
    tauPc = pub_param[1][0]
    tauQc = pub_param[1][1]
    tauPb = pub_param[1][2]
    tauQb = pub_param[1][3]
    tauPQb = pub_param[1][4]
    A2 = get_A(tauPb,tauQb,tauPQb)
    
    phiPc = Alice_pub[0]
    phiQc = Alice_pub[1]
    phiPb = Alice_pub[2]
    phiQb = Alice_pub[3]
    phiPQb = Alice_pub[4]
    A3 = get_A(phiPb,phiQb,phiPQb)


    Pb = pub_param[0][2]
    Qb = pub_param[0][3]
    PQb = pub_param[0][4]
    A1 = get_A(Pb,Qb,PQb)

    Bob_matrix = [randrange(2**(c-1)),randrange(2**(c-1))]
    Bob_matrix[0] = 2*Bob_matrix[0]+1
    Bob_matrix[1] = 2*Bob_matrix[1]+1

    Bob_secret = 5*(randrange(5**(b-1))) + (randrange(4)) + 1

    R2 = Ladder_3pt(tauPb,tauQb,tauPQb,A2,Bob_secret)
    R3 = Ladder_3pt(phiPb,phiQb,phiPQb,A3,Bob_secret)
    R1 = Ladder_3pt(Pb,Qb,PQb,A1,Bob_secret)


    AP2 = isog_5_strategy(b,A2,[tauPc,tauQc],R2,strategy)
    A2 = AP2[0]
    tauPc = AP2[1][0]
    tauQc = AP2[1][1]

    AP3 = isog_5_strategy(b,A3,[phiPc,phiQc],R3,strategy)
    A3 = AP3[0]
    phiPc = AP3[1][0]
    phiQc = AP3[1][1]

    AP1 = isog_5_strategy(b,A1,[],R1,strategy)
    A1 = AP1[0]

    tauPc = scalar(tauPc,A2,Bob_matrix[0])
    tauQc = scalar(tauQc,A2,Bob_matrix[1])
    phiPc = scalar(phiPc,A3,Bob_matrix[0])
    phiQc = scalar(phiQc,A3,Bob_matrix[1])
    
    Bob_pub = [[A2,tauPc,tauQc],[A3,phiPc,phiQc]]

    Bob_shared_key = j_inv_Mon(A1)
    
    return Bob_pub,Bob_shared_key


def Alice_shared(c,a,ell,Bob_pub,Alice_secret,p,Fp,Fp2,compressed):

    A1 = Bob_pub[1][0]
    phiPc = Bob_pub[1][1]
    phiQc = Bob_pub[1][2]
    E1,E1points = MontoEllip(A1,[phiPc,phiQc])
    
    A2 = Bob_pub[0][0]
    tauPc = Bob_pub[0][1]
    tauQc = Bob_pub[0][2]

    tauPc = scalar(tauPc,A2,Alice_secret[0])
    tauQc = scalar(tauQc,A2,Alice_secret[1])

    R = [1,0]
    while R[1] == 0:
        xR = Fp2.random_element()
        R = [xR,1]
        R = scalar(R,A2,(p+1)//ell)
        if scalar(R,A2,ell)[1] != 0:
            R[1] = 0

    kernel = []
    for i in range(((ell-1)//2)):
        kernel.append(scalar(R,A2,(i+1)))

    A2 = isog_curve(A2,kernel)
    tauPc = isog_point(tauPc,kernel)
    tauQc = isog_point(tauQc,kernel)
    E2,E2points = MontoEllip(A2,[tauPc,tauQc])

    if (E1points[0].weil_pairing(E1points[1],2**c))**(2**c-3**a) != (E2points[0].weil_pairing(E2points[1],2**c))**(3**a):
        E2points[1] = (-1)*E2points[1]
    
    strategy = optimised_strategy(c-1)
    N_constant = Fp2(2)**c/(2**c-3**a-3**a)
    
    curves,Phi = split_richelot_chain(E2points[0], E2points[1], E1points[0], E1points[1], c, N_constant, strategy,compressed)

    Alice_shared_key = [curves[0].j_invariant(),curves[1].j_invariant()]

    return Alice_shared_key



def Alice_shared_check(c,a,ell,Bob_pub,Alice_secret,p,Fp,Fp2,compressed):

    A1 = Bob_pub[1][0]
    phiPc = Bob_pub[1][1]
    phiQc = Bob_pub[1][2]
    E1,E1points = MontoEllip(A1,[phiPc,phiQc])
    
    A2 = Bob_pub[0][0]
    tauPc = Bob_pub[0][1]
    tauQc = Bob_pub[0][2]

    tauPc = scalar(tauPc,A2,Alice_secret[0])
    tauQc = scalar(tauQc,A2,Alice_secret[1])

    R = [1,0]
    R2 = [1,0]
    while (R[1] == 0) or (R2[1] == 0):
        xR = Fp2.random_element()
        R = [xR,1]
        R = scalar(R,A2,(p+1)//ell)
        xR = Fp2.random_element()
        R2 = [xR,1]
        R2 = scalar(R2,A2,(p+1)//ell)
        
        if (scalar(R,A2,ell)[1] != 0) or (scalar(R2,A2,ell)[1] != 0):
            R[1] = 0
        else:
            kernel2 = []
            RR2 = R2
            for i in range(((ell-1)//2)):
                RR2 = scalar(R2,A2,(i+1))
                kernel2.append(RR2)
                if R[1]*RR2[0] == R[0]*RR2[1]:
                    R[1] = 0

    kernel = []
    for i in range(((ell-1)//2)):
        kernel.append(scalar(R,A2,(i+1)))
    

    AA2 = isog_curve(A2,kernel)
    tauPcc = isog_point(tauPc,kernel)
    tauQcc = isog_point(tauQc,kernel)
    E2,E2points = MontoEllip(AA2,[tauPcc,tauQcc])

    if (E1points[0].weil_pairing(E1points[1],2**c))**(2**c-3**a) != (E2points[0].weil_pairing(E2points[1],2**c))**(3**a):
        E2points[1] = (-1)*E2points[1]

    AA2_ = isog_curve(A2,kernel2)
    tauPcc_ = isog_point(tauPc,kernel2)
    tauQcc_ = isog_point(tauQc,kernel2)
    E2_,E2points_ = MontoEllip(AA2_,[tauPcc_,tauQcc_])

    if (E1points[0].weil_pairing(E1points[1],2**c))**(2**c-3**a) != (E2points_[0].weil_pairing(E2points_[1],2**c))**(3**a):
        E2points_[1] = (-1)*E2points_[1]
    
    strategy = optimised_strategy(c-1)
    N_constant = Fp2(2)**c/(2**c-3**a-3**a)
    
    curves,Phi = split_richelot_chain(E2points[0], E2points[1], E1points[0], E1points[1], c, N_constant, strategy,compressed)

    curves2,Phi = split_richelot_chain(E2points_[0], E2points_[1], E1points[0], E1points[1], c, N_constant, strategy,compressed)

    
    Alice_shared_key = [curves[0].j_invariant(),curves[1].j_invariant()]
    Alice_shared_key2 = [curves2[0].j_invariant(),curves2[1].j_invariant()]
    List = list(set(Alice_shared_key) & set(Alice_shared_key2))

    return List[0]

