#Montgomery curves arithmetics

from random import randrange
from sage.all import EllipticCurve


def diffadd(P,Q,PQ):

    Sum = [0,0]
    t0 = P[0] + P[1]
    t1 = P[0] - P[1]
    t2 = Q[0] + Q[1]
    t3 = Q[0] - Q[1]
    t0 *= t3
    t1 *= t2
    Sum[0] = PQ[1]*(t0 + t1)**2
    Sum[1] = PQ[0]*(t0 - t1)**2
    return Sum

def double(P,A24):

    A = [4*A24[0] - 2*A24[1],A24[1]]
    Db = [0,0]
    
    t0 = (P[0]-P[1])**2
    t1 = (P[0]+P[1])**2
    Db[1] = 4*A[1]*t0
    Db[0] = Db[1]*t1
    t1 = t1 - t0
    t0 = (2*A[1]+A[0])*t1
    Db[1] += t0
    Db[1] *= t1

    return Db

def double_add(P,Q,PQ,A24):
    
    Sum = [0,0]
    Db = [0,0]
    
    t0 = P[0] + P[1]
    t1 = P[0] - P[1]
    t2 = Q[0] + Q[1]
    t3 = Q[0] - Q[1]
    Db[0] = t0**2
    Db[1] = t1**2
    t0 = t0 * t3
    t1 = t1 * t2
    t2 = Db[0] - Db[1]
    Db[1] *= A24[1]
    Db[0] *= Db[1]
    Sum[0] = A24[0] * t2
    Sum[1] = t0 - t1
    Db[1] += Sum[0]
    Sum[0] = t0 + t1
    Db[1] *= t2
    Sum[0] = Sum[0]**2 * PQ[1]
    Sum[1] = Sum[1]**2 * PQ[0]

    return Db,Sum


def scalar(P,A24,n):

    P0 = [P[0],P[1]]
    Q = double(P,A24)
    for i in range(n.bit_length() - 1):
        if n & (1 << (n.bit_length()-i-2)):
            Q,P = double_add(Q,P,P0,A24)
        
        else:
            P,Q = double_add(P,Q,P0,A24)

    return P


def Ladder_3pt(P,Q,PQ,A24,n):
    
    P0 = [P[0],P[1]]
    Q0 = [Q[0],Q[1]]
    PQ0 = [PQ[0],PQ[1]]

    prevbit = 0
    for i in range(n.bit_length()):
        bit =  n & (1 << i)
        swap = bit ^ prevbit
        prevbit = bit
        
        if swap == 1:
            P0 = [P[0],P[1]]
            P = [PQ[0],PQ[1]]
            PQ = [P0[0],P0[1]]

        Q,PQ = double_add(Q,PQ,P,A24)

    return PQ


def isog_point(P,G):

    phiP = [1,1]
    t0 = P[0] + P[1]
    t1 = P[0] - P[1]
    
    for i in range(len(G)):
        t2 = (G[i][0] - G[i][1]) * t0
        t3 = (G[i][0] + G[i][1]) * t1
        phiP[0] *= (t2 + t3)
        phiP[1] *= (t2 - t3)

    phiP[0] = phiP[0]**2*P[0]
    phiP[1] = phiP[1]**2*P[1]
    
    return phiP


def isog_curve(A24,G):

    ell = len(G)
    Aimage = [A24[0]**(2*ell+1),A24[1]**(2*ell+1)]

    Aimage2 = [1,1]
    for i in range(ell):
        Aimage2[0] *= (2*G[i][0])
        Aimage2[1] *= (G[i][1]-G[i][0])

    Aimage[0] *= Aimage2[0]**8
    Aimage[1] *= Aimage2[1]**8
    
    return Aimage



def get_A(P,Q,PQ):

    xP = P[0]/P[1]
    xQ = Q[0]/Q[1]
    xR = PQ[0]/PQ[1]
    A = (xR*xP+xR*xQ+xP*xQ-1)**2/(4*xP*xQ*xR)
    A += -(xP+xQ+xR)

    return [A+2,4]


def MontoEllip(A24,S):

    A = [4*A24[0] - 2*A24[1],A24[1]]
    E = EllipticCurve([0,A[0]/A[1],0,1,0])

    Points = []
    for i in range(len(S)):
        xP = S[i][0]/S[i][1]
        yP = (xP**3+xP**2*A[0]/A[1]+xP).sqrt()
        P = E(xP,yP)
        Points.append(P)

    return E,Points


def j_inv_Mon(A24):

    A = (4*A24[0] - 2*A24[1])/A24[1]
    A2 = A**2
    
    j_A = 256*(A2 - 3)**3/(A2 - 4)

    return j_A


def isog_3_random(points,kernel,zeta3,p):

    for i in range(len(points)):
        points[i] = isog_point(points[i],[[kernel,1]])

    t = kernel

    t2 = t**2
    t3 = t2 * t

    p21 = p**2 - 1
    p213 = p21//3
    if p213%3 == 1:
       alpha = (t3 - t)**((p213-1)//3)
       alpha = alpha**(-1)
    else:
       alpha = (t3 - t)**((p213+1)//3)
    # (p**2-1)/3 is coprime to 3
    
    j = randrange(3)
    alpha *= zeta3**j

    kernel = 3*t*alpha**2 + (3*t2 - 1)*alpha + 3*t3 - 2*t
    
    return points,kernel


def xQuint(P,A24):
    P0 = [P[0],P[1]]
    P2 = double(P,A24)
    P = diffadd(P2,P,P0)
    P = diffadd(P,P2,P0)

    return P

def xQuinte(P,A24,e):
    for i in range(e):
        P = xQuint(P,A24)

    return P


#The code in https://crypto.stackexchange.com/questions/58375/how-to-get-an-optimal-strategy-in-computing-isogenies-for-sidh-sike provided by Luca De Feo

def optimal_strategy_5(n, p, q):
    S = { 1: [] }
    C = { 1: 0 }
    for i in range(2, n+2):
        b, cost = min(((b, C[i-b] + C[b] + b*p + (i-b)*q) for b in range(1,i)), key=lambda t: t[1])
        S[i] = [b] + S[i-b] + S[b]
        C[i] = cost
    return S[n+1]


def isog_5_strategy(b,A24,points,P,strategy):

    S = [[b,P]]
    i = 0

    while len(S) != 0:

        hP = S.pop()
        h = hP[0]
        P = hP[1]

        if h == 1:
            kernel = [P,double(P,A24)]
            A24 = isog_curve(A24,kernel)
            S2 = []
            while len(S) != 0:
                hP = S.pop(0)
                P = isog_point(hP[1],kernel)
                S2.append([hP[0]-1,P])

            S = S2.copy()
            for j in range(len(points)):
                points[j] = isog_point(points[j],kernel)

        elif (strategy[i] < h) and (0 < strategy[i]):
            S.append([h,P])
            P = xQuinte(P,A24,strategy[i])
            S.append([h-strategy[i],P])
            i += 1

    return [A24,points]
