from random import randint
from math import floor
from sage.all import EllipticCurve
from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
from sage.schemes.elliptic_curves.hom_frobenius import EllipticCurveHom_frobenius

from cornacchia.cornacchia import Cornacchia
from montgomery_arithmetics import *
from richelot_isogeny.richelot_isogenies import *


def generate_gamma(c,a,b,ell,p,Fp,Fp2):

    tau = (2**c - 3**a)//ell
    N = tau*5**(b*2)

    m = floor((N/(p*2))**(1/2))

    List = []

    while (len(List) == 0):
        z = 0
        t = 0
        while (z%5 == 0) and (t%5 == 0):
            z = randint(-m,m)
            t = randint(-m,m)
        NN = N - p*(z**2+t**2)
        List = Cornacchia(NN,1)

    return [List[0],List[1],z,t]


def generate_publicparameters(c,a,b,ell,gamma,p,Fp,Fp2,sqrtminus):
    checker = 1
    while checker == 1:
        xP = Fp.random_element()
        Pb = [xP,1]
        Pb = scalar(Pb,[1,2],(p+1)//5**b)
        Q = scalar(Pb,[1,2],5**(b-1))
        if Q[1] != 0:
            checker = 0

    Qb = [Pb[0]*(-1),Pb[1]]

    checker = 1
    while checker == 1:
        xP = Fp2.random_element()
        Pc = [xP,1]
        Pc = scalar(Pc,[1,2],(p+1)//2**c)
        P = scalar(Pc,[1,2],2**(c-1))
        xP = Fp2.random_element()
        Qc = [xP,1]
        Qc = scalar(Qc,[1,2],(p+1)//2**c)
        Q = scalar(Qc,[1,2],2**(c-1))
        if (P[1] != 0) and (P[0] != 0) and (scalar(P,[1,2],2)[1] == 0) and (Q[1] != 0) and (Q[0] != 0) and (scalar(Q,[1,2],2)[1] == 0) and (P[0]*Q[1] != P[1]*Q[0]):
            checker = 0


    E,S = MontoEllip([Fp2(1),2],[Pb,Qb,Pc,Qc])

    pi = EllipticCurveHom_frobenius(E)

    iota = WeierstrassIsomorphism(E, (sqrtminus,0,0,0))

    gammaPb = gamma[0]*S[0]+gamma[1]*iota(S[0])+gamma[2]*pi(S[0])+gamma[3]*pi(iota(S[0]))

    gammaQb = gamma[0]*S[1]+gamma[1]*iota(S[1])+gamma[2]*pi(S[1])+gamma[3]*pi(iota(S[1]))


    gammaPc = gamma[0]*S[2]+gamma[1]*iota(S[2])+gamma[2]*pi(S[2])+gamma[3]*pi(iota(S[2]))

    gammaQc = gamma[0]*S[3]+gamma[1]*iota(S[3])+gamma[2]*pi(S[3])+gamma[3]*pi(iota(S[3]))


    if gammaQb.order() < gammaPb.order():
        s = gammaPb.discrete_log(gammaQb)
        kernel1 = S[1] - s*S[0]
    else:
        s = gammaQb.discrete_log(gammaPb)
        kernel1 = S[0] - s*S[1]
    kernel1x = [kernel1[0],kernel1[2]]

    A1 = [1,2]
    Pc1 = [Pc[0],Pc[1]]
    Qc1 = [Qc[0],Qc[1]]
    for i in range(b):
        kernel = scalar(kernel1x,A1,5**(b-i-1))
        G = [kernel,double(kernel,A1)]
        kernel1x = isog_point(kernel1x,G)
        Pc1 = isog_point(Pc1,G)
        Qc1 = isog_point(Qc1,G)
        A1 = isog_curve(A1,G)


    Pc1 = scalar(Pc1,A1,5**b)
    Qc1 = scalar(Qc1,A1,5**b)


    gammaPb2 = gamma[0]*S[0]-gamma[1]*iota(S[0])-gamma[2]*pi(S[0])-gamma[3]*pi(iota(S[0]))

    gammaQb2 = gamma[0]*S[1]-gamma[1]*iota(S[1])-gamma[2]*pi(S[1])-gamma[3]*pi(iota(S[1]))


    if gammaQb2.order() < gammaPb2.order():
        s2 = gammaPb2.discrete_log(gammaQb2)
        kernel2 = S[1] - s2*S[0]
    else:
        s2 = gammaQb2.discrete_log(gammaPb2)
        kernel2 = S[0] - s2*S[1]
    kernel2x = [kernel2[0],kernel2[2]]

    A2 = [1,2]
    Pc2 = [gammaPc[0],gammaPc[2]]
    Qc2 = [gammaQc[0],gammaQc[2]]

    for i in range(b):
        kernel = scalar(kernel2x,A2,5**(b-i-1))
        G = [kernel,double(kernel,A2)]
        kernel2x = isog_point(kernel2x,G)
        Pc2 = isog_point(Pc2,G)
        Qc2 = isog_point(Qc2,G)
        A2 = isog_curve(A2,G)


    #(A1,Pc1,Qc1) and (A2,Pc2,Qc2)


    R = [1,0]
    while R[1] == 0:
        xR = Fp2.random_element()
        R = [xR,1]
        R = scalar(R,A2,(p+1)//ell)
        if scalar(R,A2,ell)[1] != 0:
            R[1] = 0

    GR = [R]
    for i in range(((ell-1)//2) - 1):
        GR.append(scalar(R,A2,(i+2)))

    R2 = [1,0]
    while R2[1] == 0:
        xR = Fp.random_element()
        R2 = [xR,1]
        R2 = scalar(R2,A2,(p+1)//ell)
        if scalar(R2,A2,ell)[1] != 0:
            R2[1] = 0
        for i in range(len(GR)):
            if (R2[0]*GR[i][1]-R2[1]*GR[i][0]) == 0:
                R2[1] = 0


    A3 = isog_curve(A2,GR)
    Pc3 = isog_point(Pc2,GR)
    Qc3 = isog_point(Qc2,GR)
    R2 = isog_point(R2,GR)
    GR2 = [R2]
    for i in range(((ell-1)//2) - 1):
        GR2.append(scalar(R2,A3,(i+2)))

    checker = 0

    b1 = b
    if ell == 5:
        b1 = b + 1

    while checker == 0:
        xR = Fp2.random_element()
        Pb = [xR,1]
        Pb = scalar(Pb,A1,(p+1)//(5**(b1)))
        sPb = scalar(Pb,A1,5**(b1-1))
        xR2 = Fp2.random_element()
        Qb = [xR2,1]
        Qb = scalar(Qb,A1,(p+1)//(5**(b1)))
        sQb = scalar(Qb,A1,5**(b1-1))
        if (sPb[1] != 0) and (sQb[1] != 0) and (scalar(sPb,A1,5)[1] == 0) and (scalar(sQb,A1,5)[1] == 0):
            checker = 1
            for i in range(2):
                sQb = scalar(sQb,A1,(i+1))
                if (sPb[1]*sQb[0]-sPb[0]*sQb[1]) == 0:
                    checker = 0


    Es,Espoints = MontoEllip(A1,[Pb,Qb,Pc1,Qc1])
    Estilde,Estildepoints = MontoEllip(A3,[Pc3,Qc3])

    PQb_ = Espoints[0] - Espoints[1]
    PQb = [PQb_[0],PQb_[2]]

    if (Espoints[2].weil_pairing(Espoints[3],2**c))**(2**c-3**a) != (Estildepoints[0].weil_pairing(Estildepoints[1],2**c)):
        Estildepoints[1] = (-1)*Estildepoints[1]


    strategy = optimised_strategy(c-1)

    N_constant = Fp2(2)**c/(2**c-3**a-3**a)

    curves,Phi = split_richelot_chain(3**(a//2)*Espoints[2], 3**(a//2)*Espoints[3], Estildepoints[0], Estildepoints[1], c, N_constant, strategy, 0)

    PbQb = (Espoints[0],Estilde(0))
    tauPb = evaluate_richelot_chain(Phi, PbQb)[1]
    PbQb = (Espoints[1],Estilde(0))
    tauQb = evaluate_richelot_chain(Phi, PbQb)[1]


    Isom = curves[1].isomorphism_to(Estilde)

    tauPb = Isom(tauPb)
    tauQb = Isom(tauQb)

    tauPQb = tauPb - tauQb

    tauPbmon = [tauPb[0],tauPb[2]]
    tauQbmon = [tauQb[0],tauQb[2]]
    tauPQbmon = [tauPQb[0],tauPQb[2]]

    tauPbmon = isog_point(tauPbmon,GR2)
    tauQbmon = isog_point(tauQbmon,GR2)
    tauPQbmon = isog_point(tauPQbmon,GR2)


    Pb = scalar(Pb,A1,ell)
    Qb = scalar(Qb,A1,ell)
    PQb = scalar(PQb,A1,ell)

    pub_param = [[Pc1,Qc1,Pb,Qb,PQb],[Pc2,Qc2,tauPbmon,tauQbmon,tauPQbmon]]

    return pub_param
